/*
 * DBConnector
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * DBConnector.h
 *
 *  Created on: May 12, 2010
 *      Author: Timo Minartz
 */

#ifndef DBCONNECTOR_H_
#define DBCONNECTOR_H_

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

/**
 * max precision for uint64 and double values in string representation (= count chars)
 */
#define MAX_PRECISION 20

/**
 * error codes
 */
#define OK 0
#define DB_CONNECTION_FAILED -1
#define DB_QUERY_FAILED -2
#define UNEXPECTED_TABLE_DEFINITONS -3
#define UNEXPECTED_METRIC -4
#define DB_NO_SUCH_NODE -5


/**
 * Struct for one database row
 */
typedef struct {
	uint64_t timestamp; /** the timestamp in microseconds */
	double value; /** the double value for the corresponding value */
} row_struct;


/**
 * creates connection to database
 * @param bufferSize Size of the internal query buffer. bufferSize count of queries are grouped into one database query to avoid TCP overhead
 * @param conninfo http://www.postgresql.org/docs/8.4/static/libpq-connect.html
 * @return OK or in case of error DB_CONNECTION_FAILED
 */
int initConnectionWithArgs(int bufferSize, const char *conninfo);

/**
 * creates connection to database with default eeclust values, is called automatically if connection is not initialized
 * @param bufferSize Size of the internal query buffer. bufferSize count of queries are grouped into one database query to avoid TCP overhead
 * @return OK or in case of error DB_CONNECTION_FAILED
 */
int db_initConnection(int bufferSize);

/**
 * creates table for node in database
 * @warning Only for compatibility reason. Use createNodeMetric() instead
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int createNodeTable(const char * node);

/**
 * creates table for node and corresponding metric in database
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int createNodeMetric(const char * metric, const char * node);

/**
 * sets pointer to all existing nodes (= tables) in the database
 * @nodes pointer to array of strings has to be NULL, contains node names after method call
 * @return count of nodes, in case of error DB_CONNECTION_FAILED or QUERY_FAILED
 */
int getNodes(char *** nodes);

/**
 * sets pointer to all existing metrics (= tables) in the database for specified node
 * @metrics pointer to array of strings has to be NULL, contains metrics names after method call
 * @return count of metrics, in case of error DB_CONNECTION_FAILED or QUERY_FAILED
 */
int getMetrics(const char * node, char *** metrics);

/**
 * inserts timestamp and power value in the database table for node
 * @warning Only for compatibility reason. Use db_pushNodeMetric() instead
 * @param timestamp Timestamp in microseconds
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int pushNodeData(const char * node, uint64_t timestamp, double power);
void db_finishConnection(void);

/**
 * inserts timestamp and value in the database table for node depending on metric
 * @param timestamp Timestamp in microseconds
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int db_pushNodeMetric(const char * metric, const char * node, uint64_t timestamp, double value);

/**
 * gets timestamp and power value from database for the node where the timestamp is between the two specified timestamps
 * @warning Only for compatibility reason. Use getNodeMetric() instead
 * @param from_timestamp Timestamp in microseconds
 * @param to_timestamp Timestamp in microseconds
 * @rows pointer to array of table rows has to be NULL, contains table data after method call
 * @return count of rows, in case of error DB_CONNECTION_FAILED or QUERY_FAILED
 */
int getNodeData(const char * node, const uint64_t from_timestamp, const uint64_t to_timestamp, row_struct ** rows);

/**
 * gets timestamp and value from database for the node where the timestamp is between the two specified timestamps for the specified metric
 *
 * @rows pointer to array of table rows has to be NULL, contains table data after method call
 * @param from_timestamp Timestamp in microseconds
 * @param to_timestamp Timestamp in microseconds
 * @return count of rows, in case of error DB_CONNECTION_FAILED or QUERY_FAILED
 */
int db_getNodeMetric(const char * metric, const char * node, const uint64_t from_timestamp, const uint64_t to_timestamp, row_struct ** rows);

/**
 * deletes rows from database where the timestamp is between the two specified timestamps
 * @warning Only for compatibility reason. Use deleteNodeMetric() instead
 * @param from_timestamp Timestamp in microseconds
 * @param to_timestamp Timestamp in microseconds
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int deleteNodeData(const char * node, const uint64_t from_timestamp, const uint64_t to_timestamp);

/**
 * deletes rows from database where the timestamp is between the two specified timestamps for specified metric
 * @param from_timestamp Timestamp in microseconds
 * @param to_timestamp Timestamp in microseconds
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int deleteNodeMetric(const char * metric, const char * node, const uint64_t from_timestamp, const uint64_t to_timestamp);

/**
 * deletes the whole table for a node, but NOT its metrics
 * @warning Only for compatibility reason. Use deleteMetric() instead
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int deleteNodeTable(const char * node);

/**
 * deletes the whole table for a metric for a specied node
 * @return OK or in case of error DB_CONNECTION_FAILED, QUERY_FAILED
 */
int deleteMetric(const char * metric, const char * node);

/**
 * @return actual system timestamp in microseconds
 */
uint64_t getSystemTimestamp(void);


#endif /* DBCONNECTOR_H_ */
