/*
 * DBConnector
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * PostgresConnector.c
 *
 *  Created on: Apr 30, 2010
 *      Author: Timo Minartz
 */

#include "DBConnector.h"

#ifdef HAVE_PG

#include "libpq-fe.h"

#include <glib.h>

static const char * conninfo = "host = master dbname = PowerTracer user = eeclust password = eeclust";

int initConnectionWithArgs(int bufferSize, const char *new_conninfo) {

	conninfo = strdup(new_conninfo);

	return db_initConnection(bufferSize);
}

int createNodeTable(const char * node) {
	return createNodeMetric(node, "power");
}

int createNodeMetric(const char * node, const char * metric) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 100 + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query,countChars,"CREATE TABLE %s_%s ( timestamp bigint PRIMARY KEY, value double precision NOT NULL );", node, metric);

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_COMMAND_OK) {
		printf("createNodeMetric failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		free(query);
		PQclear(res);
		PQfinish(connection);
		return DB_QUERY_FAILED;
	}

	free(query);
	PQclear(res);
	PQfinish(connection);
	return OK;
}

int getNodes(char ***p_nodes) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	// returns all tablenames not containing an underscore
	const char
	* query =
	"SELECT c.relname as Tables_in FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE pg_catalog.pg_table_is_visible(c.oid) AND c.relkind='r' AND relname NOT LIKE E'%\\\\_%' ORDER BY 1;";

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_TUPLES_OK) {
		printf("getNodes failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		PQclear(res);
		PQfinish(connection);
		return DB_QUERY_FAILED;
	}

	int countRows = PQntuples(res);
	int countCols = PQnfields(res);

	if (countRows == 0) {
		PQfinish(connection);
		return countRows;
	}

	if (countCols != 1) {
		PQfinish(connection);
		return -2;
	}

	// put all nodes (= table name until first underscore) in a list, if not already in this list

	GList *nodeList = NULL;

	int i, size = 0;

	for (i = 0; i < countRows; ++i) {

		gchar ** tmp = g_strsplit(PQgetvalue(res, i, 0), "_", 2);

		if(i == 1 || strcmp(nodeList->data, tmp[0]) != 0) {
			nodeList = g_list_prepend(nodeList, strdup(tmp[0]));
			size++;
		}

		g_strfreev(tmp);
	}

	// copy list entries into array

	char ** nodes = malloc(size * sizeof(char*));

	i = 0;

	for (GList* it = nodeList; it != NULL; it = it->next) {

		nodes[i++] = (gchar *) it->data;
	}

	g_list_free(nodeList);

	*p_nodes = nodes;

	PQclear(res);

	PQfinish(connection);

	return countRows;
}

int getMetrics(const char * node, char ***p_metrics) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 300 + strlen(node);

	char * query = malloc(countChars * sizeof(char));

	snprintf(
			query,
			countChars,
			"SELECT c.relname as Tables_in FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE pg_catalog.pg_table_is_visible(c.oid) AND c.relkind='r' AND relname NOT LIKE E'pg\\\\_%%' AND relname LIKE E'%s\\\\_\%%' ORDER BY 1;",
			node);

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_TUPLES_OK) {
		printf("getMetrics failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		PQclear(res);
		PQfinish(connection);
		return DB_QUERY_FAILED;
	}

	int countRows = PQntuples(res);
	int countCols = PQnfields(res);

	if (countRows == 0) {
		PQfinish(connection);
		return countRows;
	}

	if (countCols != 1) {
		PQfinish(connection);
		return -2;
	}

	char ** metrics = malloc(countRows * sizeof(char*));

	int i;

	for (i = 0; i < countRows; ++i) {

		// remove the leading node_ from tablename
		char* metric = strdup(PQgetvalue(res, i, 0) + strlen(node) + 1);

		metrics[i] = metric;

	}

	PQclear(res);

	*p_metrics = metrics;

	PQfinish(connection);

	return countRows;
}

int pushNodeData(const char * node, uint64_t timestamp, double value) {
	return db_pushNodeMetric("power", node, timestamp, value);
}

int getNodeData(const char * node, const uint64_t from_timestamp,
		const uint64_t to_timestamp, row_struct ** rows) {
	return db_getNodeMetric("power", node, from_timestamp, to_timestamp, rows);
}

int deleteNodeData(const char * node, const uint64_t from_timestamp,
		const uint64_t to_timestamp) {
	return deleteNodeMetric("power", node, from_timestamp, to_timestamp);
}

int deleteNodeMetric(const char * metric, const char * node,
		const uint64_t from_timestamp, const uint64_t to_timestamp) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 45 + (2 * MAX_PRECISION) + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars,"DELETE FROM %s_%s WHERE timestamp BETWEEN %lu AND %lu ;",node, metric, from_timestamp, to_timestamp);

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_COMMAND_OK) {
		printf("deleteNodeMetric failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		free(query);
		PQclear(res);

		PQfinish(connection);

		return DB_QUERY_FAILED;
	}

	free(query);
	PQclear(res);

	PQfinish(connection);

	return OK;

}

int deleteNodeTable(const char * node) {
	return deleteMetric("power", node);
}

int deleteMetric(const char * metric, const char * node) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 15 + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "DROP TABLE %s_%s ;", node, metric);

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_COMMAND_OK) {
		printf("deleteNodeMetric failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		free(query);
		PQclear(res);

		PQfinish(connection);

		return DB_QUERY_FAILED;
	}

	free(query);
	PQclear(res);

	PQfinish(connection);

	return OK;
}

int db_getNodeMetric(const char * metric, const char * node,
		const uint64_t from_timestamp, const uint64_t to_timestamp,
		row_struct ** rows) {

	PGconn * connection = PQconnectdb(conninfo);

	if (PQstatus(connection) != CONNECTION_OK) {
		PQfinish(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 45 + (2 * MAX_PRECISION) + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars,"SELECT * FROM %s_%s WHERE timestamp BETWEEN %lu AND %lu ;",node, metric, from_timestamp, to_timestamp);

	PGresult *res = PQexec(connection, query);

	ExecStatusType execStatus = PQresultStatus(res);

	if (execStatus != PGRES_TUPLES_OK) {

		// construct error message for not existing node table
		int tmpChars = strlen("ERROR:  relation \"\" does not exist") + strlen(node) + 1;
		char tmp[tmpChars];
		snprintf(tmp, tmpChars,"ERROR:  relation \"%s\" does not exist", node);

		if(strncmp(tmp, PQresultErrorMessage(res), strlen(tmp)) == 0) {
			return DB_NO_SUCH_NODE;
		}

		printf("getNodeMetric failed.\n");
		printf("Query: %s\n", query);
		printf("Response:%s\n", PQresultErrorMessage(res));

		free(query);
		PQclear(res);

		PQfinish(connection);

		return DB_QUERY_FAILED;
	}

	int countRows = PQntuples(res);
	int countCols = PQnfields(res);

	if (countRows == 0) {
		PQfinish(connection);
		return countRows;
	}

	if (countCols != 2) {
		free(query);
		PQclear(res);
		PQfinish(connection);
		return UNEXPECTED_TABLE_DEFINITONS;
	}

	*rows = malloc(countRows * sizeof(**rows));

	row_struct row;

	int i;

	for (i = 0; i < countRows; ++i) {

		sscanf(PQgetvalue(res, i, 0), "%lu", &row.timestamp);
		sscanf(PQgetvalue(res, i, 1), "%lf", &row.value);

		memcpy(*rows + i, &row, sizeof(row));

	}

	free(query);
	PQclear(res);

	PQfinish(connection);

	return countRows;
}

#endif
