/*
 * DBConnector
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * QueryPuffer.c
 *
 *  Created on: May 10, 2011
 *  Author: Timo Minartz
 */

#include "QueryPuffer.h"
#include <glib.h> // for asynchronous queue and threads
#ifdef HAVE_PG
#include "libpq-fe.h" // for db connection with postgres
#elif HAVE_SQLITE
#include "sqlite3.h" // for db connection with sqlite
#endif

#include <string.h> // for strcmp
#include "DBConnector.h" // for return codes
/*
 * Asynchronous queue for sending the messages via socket
 */
#define TERMINATE_CLIENT_THREAD_MSG "terminate_now"

void processQueue(QueryPuffer * queryPuffer);

QueryPuffer* QueryPuffer_alloc(const int bufferSize, const char * conninfo) {

	// check bufferSize
	if (bufferSize < 1)
		return NULL;

	QueryPuffer * queryPuffer = g_malloc(sizeof(QueryPuffer));

	queryPuffer->bufferSize = bufferSize;
	queryPuffer->conninfo = conninfo;

	/* initialize thread safe glib */
	if (!g_thread_get_initialized())
		g_thread_init(NULL);

	// initialize asynchronous queue
	queryPuffer->queue = g_async_queue_new();

	// start thread for processing the incoming messages from program side

	queryPuffer->worker = g_thread_create((GThreadFunc) & processQueue,
			(gpointer) queryPuffer, TRUE, NULL);

	if (queryPuffer->worker == NULL) {
		g_async_queue_unref(queryPuffer->queue);
		g_free(queryPuffer);
		return NULL;
	}

	return queryPuffer;
}

void QueryPuffer_addQuery(QueryPuffer * queryPuffer, const char * metric,
		const char * node, uint64_t timestamp, double value) {

	if (queryPuffer != NULL) {

		int countChars = 30 + (2 * MAX_PRECISION) + strlen(node) + strlen(
				metric);

		gchar * queryString = malloc(countChars * sizeof(gchar));

		snprintf(queryString, countChars, "INSERT INTO %s_%s VALUES ( %lu, %lf ); ", node, metric, timestamp, value);

		g_async_queue_push(queryPuffer->queue, (gpointer) queryString);
	}
}

void QueryPuffer_unref(QueryPuffer * queryPuffer) {

	if (queryPuffer != NULL) {

		g_async_queue_push(queryPuffer->queue,
				(gpointer) TERMINATE_CLIENT_THREAD_MSG);

		g_thread_join(queryPuffer->worker);

		g_async_queue_unref(queryPuffer->queue);

		g_free(queryPuffer);

		queryPuffer = NULL;
	}
}

void processQueue(QueryPuffer * queryPuffer) {

#ifdef HAVE_PG
	PGconn * connection = PQconnectdb(queryPuffer->conninfo);
#elif HAVE_SQLITE
	sqlite3 * connection = NULL;

	if (sqlite3_open(queryPuffer->conninfo, &connection)) {
		sqlite3_close(connection);
		return;
	}
#endif

	gboolean readFromQueue = TRUE;

	while (readFromQueue) {

		int i = 0;

		int countChars = queryPuffer->bufferSize * (30 + (2 * MAX_PRECISION)
				+ strlen("intel1") + strlen("power"));

		GString * queryString = g_string_sized_new(countChars);

		while (i < queryPuffer->bufferSize) {

			gchar * query = (gchar *) g_async_queue_pop(queryPuffer->queue);

			if (strcmp(query, TERMINATE_CLIENT_THREAD_MSG) == 0) {
				readFromQueue = FALSE;
				break;
			}

#ifdef DEBUG
			g_print("query from queue: %s\n", query);
#endif
			g_string_append(queryString, query);

			g_free(query);
			i++;
		}

		if (strcmp(queryString->str, "") == 0) {
			g_string_free(queryString, TRUE);
			break;
		}

#ifdef DEBUG
		g_print("Query: %s\n", queryString->str);
#endif

#ifdef HAVE_PG
		PGresult *res = PQexec(connection, queryString->str);

		ExecStatusType execStatus = PQresultStatus(res);

		if (execStatus != PGRES_COMMAND_OK) {
			printf("pushNodeMetric failed.\n");
			printf("Query: %s\n", queryString->str);
			printf("Response:%s\n", PQresultErrorMessage(res));

			g_string_free(queryString, TRUE);
			PQclear(res);

			break;
		}

		PQclear(res);

		g_string_free(queryString, TRUE);
#elif HAVE_SQLITE

		char *zErrMsg = NULL;

		int rc = sqlite3_exec(connection, queryString->str, NULL, NULL, &zErrMsg);

		if (rc != SQLITE_OK) {
			printf("pushNodeMetric failed.\n");
			printf("Query: %s\n", queryString->str);
			printf("Response:%s\n", zErrMsg);
			sqlite3_free(zErrMsg);
			g_string_free(queryString, TRUE);
			sqlite3_close(connection);

			break;
		}

		g_string_free(queryString, TRUE);

#endif

	}

#ifdef HAVE_PG
	PQfinish(connection);
#elif HAVE_SQLITE
	sqlite3_close(connection);
#endif
}
