/*
 * DBConnector
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * SQLiteConnector.c
 *
 *  Created on: May 12, 2010
 *      Author: Timo Minartz
 */

#include "DBConnector.h"
#include <glib.h>

#ifdef HAVE_SQLITE

#include "sqlite3.h"

sqlite3 * connection = NULL;

static const char * dbname = "PowerTracer";
static row_struct * rows;

int initConnectionWithArgs(int bufferSize, const char *new_dbname) {

	dbname = strdup(new_dbname);

	return db_initConnection(bufferSize);
}

int createNodeTable(const char * node) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 100 + strlen(node);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "CREATE TABLE %s ( timestamp bigint PRIMARY KEY, power double precision NOT NULL );", node);

	char *zErrMsg = NULL;

	int rc = sqlite3_exec(connection, query, NULL, NULL, &zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while creating table for node %s: %s\n", node, zErrMsg);
		sqlite3_free(zErrMsg);
		free(query);
		sqlite3_close(connection);
		return DB_QUERY_FAILED;
	}

	free(query);
	sqlite3_close(connection);
	return OK;
}

int createNodeMetric(const char * metric, const char * node) {
	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 100 + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "CREATE TABLE %s_%s ( timestamp bigint PRIMARY KEY, power double precision NOT NULL );", node, metric);

	char *zErrMsg = NULL;

	int rc = sqlite3_exec(connection, query, NULL, NULL, &zErrMsg);

	if (rc != SQLITE_OK) {

		if(g_str_has_suffix(zErrMsg, " already exists")) {
			free(query);
			sqlite3_close(connection);
			return OK;
		}

		printf("SQL error while creating table for node metric %s_%s: %s\n", node, metric, zErrMsg);
		sqlite3_free(zErrMsg);
		free(query);
		sqlite3_close(connection);
		return DB_QUERY_FAILED;
	}

	free(query);
	sqlite3_close(connection);
	return OK;
}

int getNodes(char ***p_nodes) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	// get sql mastertable containing all tables in this database
	char * query = "SELECT * FROM sqlite_master WHERE type = \"table\";";

	int countRows = 0;
	int countCols = 0;
	char** res = NULL;
	char *zErrMsg = NULL;

	int rc = sqlite3_get_table(connection,query,&res,&countRows,&countCols,&zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while getting table names: %s\n",zErrMsg);

		sqlite3_free(zErrMsg);
		sqlite3_free_table(res);
		sqlite3_close(connection);

		return DB_QUERY_FAILED;
	}

	// if there is only one row, its only the title row...
	if (countCols == 0) {
		sqlite3_free_table(res);
		sqlite3_close(connection);
		return 0;
	}

	// check standard table definitions
	if (countCols != 5) {
		sqlite3_free_table(res);
		sqlite3_close(connection);
		return UNEXPECTED_TABLE_DEFINITONS;
	}

	// put all nodes (= table name until first underscore) in a list, if not already in this list

	GList *nodeList = NULL;

	int i, size = 0;

	for (i = 1; i < countRows; ++i) {

		gchar ** tmp = g_strsplit(res[i * countCols + 1], "_", 2);

		if(i == 1 || strcmp(nodeList->data, tmp[0]) != 0) {
			nodeList = g_list_prepend(nodeList, strdup(tmp[0]));
			size++;
		}

		g_strfreev(tmp);
	}

	// copy list entries into array

	char ** nodes = malloc(size * sizeof(char*));

	i = 0;

	for (GList* it = nodeList; it != NULL; it = it->next) {

		nodes[i++] = (gchar *) it->data;
	}

	g_list_free(nodeList);

	*p_nodes = nodes;

	sqlite3_free_table(res);

	sqlite3_close(connection);

	return size;

}

int getMetrics(const char * node, char ***p_metrics) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 100 + strlen(node);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "SELECT * FROM sqlite_master WHERE type = \"table\" AND name LIKE \"\%%s\%\";", node);

	int countRows = 0;
	int countCols = 0;
	char** res = NULL;
	char *zErrMsg = NULL;

	int rc = sqlite3_get_table(connection,query,&res,&countRows,&countCols,&zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while getting metrics for node %s: %s\n",node, zErrMsg);

		sqlite3_free(zErrMsg);
		sqlite3_free_table(res);
		sqlite3_close(connection);

		return DB_QUERY_FAILED;
	}

	char ** metrics = malloc(countRows * sizeof(char*));

	int i;

	for (i = 1; i < countRows + 1; ++i) {

		// remove the leading node_ from tablename
		char* metric = strdup(res[i * countCols + 1] + strlen(node) + 1);

		metrics[i-1] = metric;
	}

	*p_metrics = metrics;

	sqlite3_free_table(res);

	sqlite3_close(connection);

	return countRows;
}

int pushNodeData(const char * node, uint64_t timestamp, double value) {
	return db_pushNodeMetric("power", node, timestamp, value);
}

int getNodeData(const char * node, const uint64_t from_timestamp, const uint64_t to_timestamp, row_struct ** rows) {
	return db_getNodeMetric("power", node, from_timestamp, to_timestamp, rows);
}

int db_getNodeMetric(const char * metric, const char * node, const uint64_t from_timestamp,
		const uint64_t to_timestamp, row_struct ** rows) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 45 + (2*MAX_PRECISION) + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "SELECT * FROM %s_%s WHERE timestamp BETWEEN %lu AND %lu ;",node, metric, from_timestamp, to_timestamp);

	int countRows = 0;
	int countCols = 0;
	char** res = NULL;
	char *zErrMsg = NULL;

	int rc = sqlite3_get_table(connection,query,&res,&countRows,&countCols,&zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while getting data for node %s: %s\n", node, zErrMsg);

		sqlite3_free(zErrMsg);
		sqlite3_free_table(res);
		free(query);
		sqlite3_close(connection);

		return DB_QUERY_FAILED;
	}

	if (countRows == 0) {
		sqlite3_free_table(res);
		free(query);
		sqlite3_close(connection);
		return countRows;
	}

	if (countCols != 2) {
		sqlite3_free_table(res);
		free(query);
		sqlite3_close(connection);
		return UNEXPECTED_TABLE_DEFINITONS;
	}

	*rows = malloc(countRows * sizeof(**rows));

	row_struct row;

	int i,j;

	for (i = 0; i < countRows; ++i) {
		sscanf(res[i * countCols + 0 + 2], "%lu", &row.timestamp);
		sscanf(res[i * countCols + 1 + 2], "%lf", &row.value);

		memcpy(*rows + i, &row, sizeof(row));
	}

	free(query);
	sqlite3_free_table(res);

	sqlite3_close(connection);

	return countRows;
}

int deleteNodeData(const char * node, const uint64_t from_timestamp,
		const uint64_t to_timestamp) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 45 + (2*MAX_PRECISION) + strlen(node);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "DELETE FROM %s WHERE timestamp BETWEEN %lu AND %lu ;", node, from_timestamp, to_timestamp);

	char *zErrMsg = NULL;

	int rc = sqlite3_exec(connection, query, NULL, NULL, &zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while deleting data for node %s: %s\n", node, zErrMsg);
		sqlite3_free(zErrMsg);
		free(query);
		sqlite3_close(connection);

		return DB_QUERY_FAILED;
	}

	free(query);
	sqlite3_close(connection);

	return OK;
}

int deleteNodeTable(const char * node) {
	deleteMetric("power", node);
}

int deleteMetric(const char * metric, const char * node) {

	sqlite3 * connection = NULL;

	if (sqlite3_open(dbname, &connection)) {
		sqlite3_close(connection);
		return DB_CONNECTION_FAILED;
	}

	int countChars = 15 + strlen(node) + strlen(metric);

	char * query = malloc(countChars * sizeof(char));

	snprintf(query, countChars, "DROP TABLE %s_%s ;", node, metric);

	char *zErrMsg = NULL;

	int rc = sqlite3_exec(connection, query, NULL, NULL, &zErrMsg);

	if (rc != SQLITE_OK) {
		printf("SQL error while deleting table for node metric %s_%s: %s\n", node, metric, zErrMsg);
		sqlite3_free(zErrMsg);
		free(query);
		sqlite3_close(connection);

		return DB_QUERY_FAILED;
	}

	free(query);
	sqlite3_close(connection);

	return OK;
}
#endif

